import torch
from torch_geometric.nn import GINEConv, global_add_pool, global_mean_pool, global_max_pool
from torch_geometric.nn.aggr import AttentionalAggregation, Set2Set
from torch.functional import F
from deepchem.data import Dataset
from deepchem.models.losses import SoftmaxCrossEntropy, EdgePredictionLoss
from deepchem.models.torch_models import ModularTorchModel
from deepchem.feat.graph_data import BatchGraphData
from typing import Iterable, List, Tuple
from deepchem.metrics import to_one_hot

num_atom_type = 120
num_chirality_tag = 3
# Relevant in future PRs
# num_bond_type = 6
# num_bond_direction = 3


class GNN(torch.nn.Module):
    """
    GNN module for the GNNModular model.

    This module is responsible for the graph neural network layers in the GNNModular model.

    Parameters
    ----------
    atom_type_embedding: torch.nn.Embedding
        Embedding layer for atom types.
    chirality_embedding: torch.nn.Embedding
        Embedding layer for chirality tags.
    gconvs: torch.nn.ModuleList
        ModuleList of graph convolutional layers.
    batch_norms: torch.nn.ModuleList
        ModuleList of batch normalization layers.
    dropout: int
        Dropout probability.
    jump_knowledge: str
        The type of jump knowledge to use. [1] Must be one of "last", "sum", "max", "concat" or "none".
        "last": Use the node representation from the last GNN layer.
        "concat": Concatenate the node representations from all GNN layers.
        "max": Take the element-wise maximum of the node representations from all GNN layers.
        "sum": Take the element-wise sum of the node representations from all GNN layers.
    init_emb: bool
        Whether to initialize the embedding layers with Xavier uniform initialization.

    References
    ----------
    .. [1] Xu, K. et al. Representation Learning on Graphs with Jumping Knowledge Networks. Preprint at https://doi.org/10.48550/arXiv.1806.03536 (2018).

    Example
    -------
    >>> from deepchem.models.torch_models.gnn import GNNModular
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> from deepchem.feat.molecule_featurizers import SNAPFeaturizer
    >>> featurizer = SNAPFeaturizer()
    >>> smiles = ["C1=CC=CC=C1", "C1=CC=CC=C1C=O", "C1=CC=CC=C1C(=O)O"]
    >>> features = featurizer.featurize(smiles)
    >>> batched_graph = BatchGraphData(features).numpy_to_torch(device="cuda")
    >>> modular = GNNModular(emb_dim = 8, task = "edge_pred")
    >>> gnnmodel = modular.gnn
    >>> print(gnnmodel(batched_graph)[0].shape)
    torch.Size([23, 32])

    """

    def __init__(self,
                 atom_type_embedding,
                 chirality_embedding,
                 gconvs,
                 batch_norms,
                 dropout,
                 jump_knowledge,
                 init_emb=False):
        super(GNN, self).__init__()

        self.atom_type_embedding = atom_type_embedding
        self.chirality_embedding = chirality_embedding
        self.gconv = gconvs
        self.batch_norms = batch_norms
        self.dropout = dropout
        self.num_layer = len(gconvs)
        self.jump_knowledge = jump_knowledge

        # may mess with loading pretrained weights
        if init_emb:
            torch.nn.init.xavier_uniform_(self.atom_type_embedding.weight.data)
            torch.nn.init.xavier_uniform_(self.chirality_embedding.weight.data)

    def forward(self, data: BatchGraphData):
        """
        Forward pass for the GNN module.

        Parameters
        ----------
        data: BatchGraphData
            Batched graph data.
        """

        node_feats = data.node_features[:, 0].long()  # type: ignore
        chiral_feats = data.node_features[:, 1].long()  # type: ignore
        x = self.atom_type_embedding(node_feats) + self.chirality_embedding(
            chiral_feats)

        h_list = [x]
        for i, conv_layer in enumerate(self.gconv):
            h = conv_layer(h_list[i], data.edge_index, data.edge_features)
            h = self.batch_norms[i](h)
            h = F.dropout(F.relu(h), self.dropout, training=self.training)
            if i == self.num_layer - 1:
                # remove relu for the last layer
                h = F.dropout(h, self.dropout, training=self.training)
            else:
                h = F.dropout(F.relu(h), self.dropout, training=self.training)
            h_list.append(h)

        # Different implementations of jump_knowledge
        if self.jump_knowledge == "concat":
            node_representation = torch.cat(h_list, dim=1)
        elif self.jump_knowledge == "last":
            node_representation = h_list[-1]
        elif self.jump_knowledge == "max":
            h_list = [h.unsqueeze_(0) for h in h_list]
            node_representation = torch.max(torch.cat(h_list, dim=0), dim=0)[0]
        elif self.jump_knowledge == "sum":
            h_list = [h.unsqueeze_(0) for h in h_list]
            node_representation = torch.sum(torch.cat(h_list, dim=0), dim=0)[0]

        return (node_representation, data)


class GNNHead(torch.nn.Module):
    """
    Prediction head module for the GNNModular model.

    Parameters
    ----------
    pool: Union[function,torch.nn.Module]
        Pooling function or nn.Module to use
    head: torch.nn.Module
        Prediction head to use
    task: str
        The type of task. Must be one of "regression", "classification".
    num_tasks: int
        Number of tasks.
    num_classes: int
        Number of classes for classification.
    """

    def __init__(self, pool, head, task, num_tasks, num_classes):
        super().__init__()
        self.pool = pool
        self.head = head
        self.task = task
        self.num_tasks = num_tasks
        self.num_classes = num_classes

    def forward(self, data):
        """
        Forward pass for the GNN head module.

        Parameters
        ----------
        data: tuple
            A tuple containing the node representations and the input graph data.
            node_representation is a torch.Tensor created after passing input through the GNN layers.
            input_batch is the original input BatchGraphData.
        """
        node_representation, input_batch = data

        pooled = self.pool(node_representation, input_batch.graph_index)
        out = self.head(pooled)
        if self.task == "classification":
            out = torch.reshape(out, (-1, self.num_tasks, self.num_classes))
        return out


class GNNModular(ModularTorchModel):
    """
    Modular GNN which allows for easy swapping of GNN layers.

    Parameters
    ----------
    gnn_type: str
        The type of GNN layer to use. Must be one of "gin", "gcn", "graphsage", or "gat".
    num_layer: int
        The number of GNN layers to use.
    emb_dim: int
        The dimensionality of the node embeddings.
    num_tasks: int
        The number of tasks.
    graph_pooling: str
        The type of graph pooling to use. Must be one of "sum", "mean", "max", "attention" or "set2set".
    dropout: float, optional (default 0)
        The dropout probability.
    jump_knowledge: str, optional (default "last")
        The type of jump knowledge to use. [1] Must be one of "last", "sum", "max", "concat" or "none".
        "last": Use the node representation from the last GNN layer.
        "concat": Concatenate the node representations from all GNN layers.
        "max": Take the element-wise maximum of the node representations from all GNN layers.
        "sum": Take the element-wise sum of the node representations from all GNN layers.
    task: str, optional (default "regression")
        The type of task. Can be unsupervised tasks "edge_pred" or "node_pred" or supervised tasks like "regression" or "classification".

    Examples
    --------
    >>> import numpy as np
    >>> import deepchem as dc
    >>> from deepchem.feat.molecule_featurizers import SNAPFeaturizer
    >>> from deepchem.models.torch_models.gnn import GNNModular
    >>> featurizer = SNAPFeaturizer()
    >>> smiles = ["C1=CC=CC=C1", "C1=CC=CC=C1C=O", "C1=CC=CC=C1C(=O)O"]
    >>> features = featurizer.featurize(smiles)
    >>> dataset = dc.data.NumpyDataset(features, np.zeros(len(features)))
    >>> model = GNNModular(task="edge_pred")
    >>> loss = model.fit(dataset, nb_epoch=1)

    References
    ----------
    .. [1] Xu, K. et al. Representation Learning on Graphs with Jumping Knowledge Networks. Preprint at https://doi.org/10.48550/arXiv.1806.03536 (2018).
    .. [2] Hu, W. et al. Strategies for Pre-training Graph Neural Networks. Preprint at https://doi.org/10.48550/arXiv.1905.12265 (2020).
    """

    def __init__(self,
                 gnn_type: str = "gin",
                 num_layer: int = 3,
                 emb_dim: int = 64,
                 num_tasks: int = 1,
                 num_classes: int = 2,
                 graph_pooling: str = "attention",
                 dropout: int = 0,
                 jump_knowledge: str = "concat",
                 task: str = "edge_pred",
                 **kwargs):
        self.gnn_type = gnn_type
        self.num_layer = num_layer
        self.emb_dim = emb_dim

        self.num_tasks = num_tasks
        self.num_classes = num_classes
        if task == "classification":
            self.output_dim = num_classes * num_tasks
            self.criterion = SoftmaxCrossEntropy()._create_pytorch_loss()
        elif task == "regression":
            self.output_dim = num_tasks
            self.criterion = F.mse_loss
        elif task == "edge_pred":
            self.output_dim = num_tasks
            self.edge_pred_loss = EdgePredictionLoss()._create_pytorch_loss()

        self.graph_pooling = graph_pooling
        self.dropout = dropout
        self.jump_knowledge = jump_knowledge
        self.task = task

        self.components = self.build_components()
        self.model = self.build_model()
        super().__init__(self.model, self.components, **kwargs)

    def build_components(self):
        """
        Builds the components of the GNNModular model. It initializes the encoders, batch normalization layers, pooling layers, and head layers based on the provided configuration. The method returns a dictionary containing the following components:

        Components list, type and description:
        --------------------------------------
        atom_type_embedding: torch.nn.Embedding, an embedding layer for atom types.

        chirality_embedding: torch.nn.Embedding, an embedding layer for chirality tags.

        gconvs: torch_geometric.nn.conv.MessagePassing, a list of graph convolutional layers (encoders) based on the specified GNN type (GIN, GCN, or GAT).

        batch_norms: torch.nn.BatchNorm1d, a list of batch normalization layers corresponding to the encoders.

        pool: Union[function,torch_geometric.nn.aggr.Aggregation], a pooling layer based on the specified graph pooling type (sum, mean, max, attention, or set2set).

        head: nn.Linear, a linear layer for the head of the model.

        These components are then used to construct the GNN and GNN_head modules for the GNNModular model.
        """
        encoders = []
        batch_norms = []
        for layer in range(self.num_layer):
            if self.gnn_type == "gin":
                encoders.append(
                    GINEConv(
                        torch.nn.Linear(self.emb_dim, self.emb_dim),
                        edge_dim=2,  # bond type, bond direction
                        aggr="add"))
            else:
                raise ValueError("Only GIN is supported for now")
            # Relevent for future PRs
            # elif self.gnn_type == "gcn":
            #     encoders.append(GCNConv(self.emb_dim))
            # elif self.gnn_type == "gat":
            #     encoders.append(GATConv(self.emb_dim))
            batch_norms.append(torch.nn.BatchNorm1d(self.emb_dim))
        encoders = torch.nn.ModuleList(encoders)
        batch_norms = torch.nn.ModuleList(batch_norms)

        if self.graph_pooling == "sum":
            pool = global_add_pool
        elif self.graph_pooling == "mean":
            pool = global_mean_pool
        elif self.graph_pooling == "max":
            pool = global_max_pool
        elif self.graph_pooling == "attention":
            if self.jump_knowledge == "concat":
                pool = AttentionalAggregation(
                    gate_nn=torch.nn.Linear((self.num_layer + 1) *
                                            self.emb_dim, 1))
            else:
                pool = AttentionalAggregation(
                    gate_nn=torch.nn.Linear(self.emb_dim, 1))
        elif self.graph_pooling == "set2set":
            set2setiter = 3
            if self.jump_knowledge == "concat":
                pool = Set2Set((self.num_layer + 1) * self.emb_dim, set2setiter)
            else:
                pool = Set2Set(self.emb_dim, processing_steps=set2setiter)

        if self.graph_pooling == "set2set":
            mult = 2
        else:
            mult = 1

        if self.jump_knowledge == "concat":
            head = torch.nn.Linear(mult * (self.num_layer + 1) * self.emb_dim,
                                   self.output_dim)
        else:
            head = torch.nn.Linear(mult * self.emb_dim, self.output_dim)

        components = {
            'atom_type_embedding':
                torch.nn.Embedding(num_atom_type, self.emb_dim),
            'chirality_embedding':
                torch.nn.Embedding(num_chirality_tag, self.emb_dim),
            'gconvs':
                encoders,
            'batch_norms':
                batch_norms,
            'pool':
                pool,
            'head':
                head
        }
        self.gnn = GNN(components['atom_type_embedding'],
                       components['chirality_embedding'], components['gconvs'],
                       components['batch_norms'], self.dropout,
                       self.jump_knowledge)
        self.gnn_head = GNNHead(components['pool'], components['head'],
                                self.task, self.num_tasks, self.num_classes)
        return components

    def build_model(self):
        """
        Builds the appropriate model based on the specified task.

        For the edge prediction task, the model is simply the GNN module because it is an unsupervised task and does not require a prediction head.

        Supervised tasks such as node classification and graph regression require a prediction head, so the model is a sequential module consisting of the GNN module followed by the GNN_head module.
        """

        if self.task == "edge_pred":  # unsupervised task, does not need pred head
            return self.gnn
        elif self.task in ("regression", "classification"):
            return torch.nn.Sequential(self.gnn, self.gnn_head)
        else:
            raise ValueError(f"Task {self.task} is not supported.")

    def loss_func(self, inputs, labels, weights):
        """
        The loss function executed in the training loop, which is based on the specified task.
        """
        if self.task == "edge_pred":
            node_emb, inputs = self.model(inputs)
            loss = self.edge_pred_loss(node_emb, inputs)
        elif self.task == "regression":
            loss = self.regression_loss(inputs, labels)
        elif self.task == "classification":
            loss = self.classification_loss(inputs, labels)
        return (loss * weights).mean()

    def regression_loss(self, inputs, labels):
        out = self.model(inputs)
        reg_loss = self.criterion(out, labels)
        return reg_loss

    def classification_loss(self, inputs, labels):
        out = self.model(inputs)
        out = F.softmax(out, dim=2)
        class_loss = self.criterion(out, labels)
        return class_loss

    def _prepare_batch(self, batch):
        """
        Prepares the batch for the model by converting the GraphData numpy arrays to BatchedGraphData torch tensors and moving them to the device, then transforming the input to the appropriate format for the task.

        Parameters
        ----------
        batch: tuple
            A tuple containing the inputs, labels, and weights for the batch.

        Returns
        -------
        inputs: BatchGraphData
            The inputs for the batch, converted to a BatchGraphData object, moved to the device, and transformed to the appropriate format for the task.
        labels: torch.Tensor
            The labels for the batch, moved to the device.
        weights: torch.Tensor
            The weights for the batch, moved to the device.
        """
        inputs, labels, weights = batch
        inputs = BatchGraphData(inputs[0]).numpy_to_torch(self.device)
        if self.task == "edge_pred":
            inputs = negative_edge_sampler(inputs)

        _, labels, weights = super()._prepare_batch(([], labels, weights))

        if (len(labels) != 0) and (len(weights) != 0):
            labels = labels[0]
            weights = weights[0]

        return inputs, labels, weights

    def default_generator(
            self,
            dataset: Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        """
        This default generator is modified from the default generator in dc.models.tensorgraph.tensor_graph.py to support multitask classification. If the task is classification, the labels y_b are converted to a one-hot encoding and reshaped according to the number of tasks and classes.
        """

        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if self.task == 'classification' and y_b is not None:
                    y_b = to_one_hot(y_b.flatten(), self.num_classes).reshape(
                        -1, self.num_tasks, self.num_classes)
                yield ([X_b], [y_b], [w_b])


def negative_edge_sampler(data: BatchGraphData):
    """
    NegativeEdge is a function that adds negative edges to the input graph data. It randomly samples negative edges (edges that do not exist in the original graph) and adds them to the input graph data.
    The number of negative edges added is equal to half the number of edges in the original graph. This is useful for tasks like edge prediction, where the model needs to learn to differentiate between existing and non-existing edges.

    Parameters
    ----------
    data: dc.feat.graph_data.BatchGraphData
        The input graph data.

    Returns
    -------
    BatchGraphData
        A new BatchGraphData object with the additional attribute `negative_edge_index`.

    Example
    -------
    >>> import numpy as np
    >>> import torch
    >>> from deepchem.feat.graph_data import BatchGraphData, GraphData
    >>> from deepchem.models.torch_models.gnn import negative_edge_sampler
    >>> num_nodes_list, num_edge_list = [3, 4, 5], [2, 4, 5]
    >>> num_node_features, num_edge_features = 32, 32
    >>> edge_index_list = [
    ...     np.array([[0, 1], [1, 2]]),
    ...     np.array([[0, 1, 2, 3], [1, 2, 0, 2]]),
    ...     np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]]),
    ... ]
    >>> graph_list = [
    ...     GraphData(node_features=np.random.random_sample(
    ...         (num_nodes_list[i], num_node_features)),
    ...               edge_index=edge_index_list[i],
    ...               edge_features=np.random.random_sample(
    ...                   (num_edge_list[i], num_edge_features)),
    ...               node_pos_features=None) for i in range(len(num_edge_list))
    ... ]
    >>> batched_graph = BatchGraphData(graph_list)
    >>> batched_graph = batched_graph.numpy_to_torch()
    >>> neg_sampled = negative_edge_sampler(batched_graph)
    """
    import torch

    num_nodes = data.num_nodes
    num_edges = data.num_edges

    edge_set = set([
        str(data.edge_index[0, i].cpu().item()) + "," +
        str(data.edge_index[1, i].cpu().item())
        for i in range(data.edge_index.shape[1])
    ])

    redandunt_sample = torch.randint(0, num_nodes, (2, 5 * num_edges))
    sampled_ind = []
    sampled_edge_set = set([])
    for i in range(5 * num_edges):
        node1 = redandunt_sample[0, i].cpu().item()
        node2 = redandunt_sample[1, i].cpu().item()
        edge_str = str(node1) + "," + str(node2)
        if edge_str not in edge_set and edge_str not in sampled_edge_set and not node1 == node2:
            sampled_edge_set.add(edge_str)
            sampled_ind.append(i)
        if len(sampled_ind) == num_edges / 2:
            break

    data.negative_edge_index = redandunt_sample[:, sampled_ind]  # type: ignore

    return data



from unittest import TestCase
import numpy as np

import deepchem.rl.envs.tictactoe


class TestTicTacToeEnvironment(TestCase):

    def test_constructor(self):
        board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
        assert len(board.state) == 1
        assert board.state[0].shape == (3, 3, 2)
        assert np.sum(board.state[0]) == 1 or np.sum(board.state[0]) == 0

    def test_step(self):
        board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
        X = deepchem.rl.envs.tictactoe.TicTacToeEnvironment.X
        board._state = [np.zeros(shape=(3, 3, 2), dtype=np.float32)]
        board.step(0)
        assert np.all(board.state[0][0][0] == X)

    def test_winner(self):
        board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
        X = deepchem.rl.envs.tictactoe.TicTacToeEnvironment.X
        board.state[0][0][0] = X
        board.state[0][0][1] = X
        assert not board.check_winner(X)
        board.state[0][0][2] = X
        assert board.check_winner(X)

    def test_game_over(self):
        board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
        X = deepchem.rl.envs.tictactoe.TicTacToeEnvironment.X
        board.state[0][0][0] = X
        board.state[0][0][1] = X
        assert not board.check_winner(X)
        board.state[0][0][2] = X
        assert board.check_winner(X)

    def test_display(self):
        board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
        s = board.display()
        assert s.find("X") == -1

    def test_get_O_move(self):
        board = deepchem.rl.envs.tictactoe.TicTacToeEnvironment()
        empty = deepchem.rl.envs.tictactoe.TicTacToeEnvironment.EMPTY
        move = board.get_O_move()
        assert np.all(board.state[0][move[0]][move[1]] == empty)



import numpy as np

from deepchem.utils.typing import RDKitMol
from deepchem.feat.base_classes import MolecularFeaturizer


class PubChemFingerprint(MolecularFeaturizer):
    """PubChem Fingerprint.

    The PubChem fingerprint is a 881 bit structural key,
    which is used by PubChem for similarity searching.
    Please confirm the details in [1]_.

    References
    ----------
    .. [1] ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.pdf

    Note
    -----
    This class requires RDKit and PubChemPy to be installed.
    PubChemPy use REST API to get the fingerprint, so you need the internet access.

    Examples
    --------
    >>> import deepchem as dc
    >>> smiles = ['CCC']
    >>> featurizer = dc.feat.PubChemFingerprint()
    >>> features = featurizer.featurize(smiles)
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (881,)

    """

    def __init__(self):
        """Initialize this featurizer."""
        try:
            from rdkit import Chem  # noqa
            import pubchempy as pcp  # noqa
        except ModuleNotFoundError:
            raise ImportError("This class requires PubChemPy to be installed.")

        self.get_pubchem_compounds = pcp.get_compounds

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
        Calculate PubChem fingerprint.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            1D array of RDKit descriptors for `mol`. The length is 881.

        """
        try:
            from rdkit import Chem
            import pubchempy as pcp
        except ModuleNotFoundError:
            raise ImportError("This class requires PubChemPy to be installed.")
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )

        smiles = Chem.MolToSmiles(datapoint)
        pubchem_compound = pcp.get_compounds(smiles, 'smiles')[0]
        feature = [int(bit) for bit in pubchem_compound.cactvs_fingerprint]
        return np.asarray(feature)



import json
from typing import Dict, Optional
from collections import Counter
from rdkit import Chem
from deepchem.data import Dataset
from deepchem.feat.base_classes import Featurizer
from deepchem.utils.typing import RDKitMol, RDKitAtom, RDKitBond
from deepchem.feat.vocabulary_builders.vocabulary_builder import VocabularyBuilder


class GroverAtomVocabularyBuilder(VocabularyBuilder):
    """Atom Vocabulary Builder for Grover

    This module can be used to generate atom vocabulary from SMILES strings for
    the GROVER pretraining task. For each atom in a molecule, the vocabulary context is the
    node-edge-count of the atom where node is the neighboring atom, edge is the type of bond (single
    bond or double bound) and count is the number of such node-edge pairs for the atom in its
    neighborhood. For example, for the molecule 'CC(=O)C', the context of the first carbon atom is
    C-SINGLE1 because it's neighbor is C atom, the type of bond is SINGLE bond and the count of such
    bonds is 1. The context of the second carbon atom is C-SINGLE2 and O-DOUBLE1 because
    it is connected to two carbon atoms by a single bond and 1 O atom by a double bond.
    The vocabulary of an atom is then computed as the `atom-symbol_contexts` where the contexts
    are sorted in alphabetical order when there are multiple contexts. For example, the
    vocabulary of second C is `C_C-SINGLE2_O-DOUBLE1`. The algorithm enumerates vocabulary of all atoms
    in the dataset and makes a vocabulary to index mapping by sorting the vocabulary
    by frequency and then alphabetically.

    The algorithm enumerates vocabulary of all atoms in the dataset and makes a vocabulary to
    index mapping by sorting the vocabulary by frequency and then alphabetically. The `max_size`
    parameter can be used for setting the size of the vocabulary. When this parameter is set,
    the algorithm stops adding new words to the index when the vocabulary size reaches `max_size`.

    Parameters
    ----------
    max_size: int (optional)
        Maximum size of vocabulary

    Example
    -------
    >>> import tempfile
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> file = tempfile.NamedTemporaryFile()
    >>> dataset = dc.data.NumpyDataset(X=[['CCC'], ['CC(=O)C']])
    >>> vocab = GroverAtomVocabularyBuilder()
    >>> vocab.build(dataset)
    >>> vocab.stoi
    {'<pad>': 0, '<other>': 1, 'C_C-SINGLE1': 2, 'C_C-SINGLE2': 3, 'C_C-SINGLE2_O-DOUBLE1': 4, 'O_C-DOUBLE1': 5}
    >>> vocab.save(file.name)
    >>> loaded_vocab = GroverAtomVocabularyBuilder.load(file.name)
    >>> mol = Chem.MolFromSmiles('CC(=O)C')
    >>> loaded_vocab.encode(mol, mol.GetAtomWithIdx(1))
    4

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """

    def __init__(self, max_size: Optional[int] = None):
        self.specials = ('<pad>', '<other>')
        self.min_freq = 1
        self.size = max_size
        self.itos = list(self.specials)
        self.stoi = self._make_reverse_mapping(self.itos)
        self.pad_index = 0
        self.other_index = 1

    def build(self, dataset: Dataset) -> None:
        """Builds vocabulary

        Parameters
        ----------
        dataset: dc.data.Dataset
            A dataset object with SMILEs strings in X attribute.
        """
        counter: Dict[str, int] = Counter()
        for x, _, _, _ in dataset.itersamples():
            smiles = x[0]
            mol = Chem.MolFromSmiles(smiles)
            for atom in mol.GetAtoms():
                v = self.atom_to_vocab(mol, atom)
                counter[v] += 1

        # sort first by frequency, then alphabetically
        words_and_frequencies = sorted(counter.items(), key=lambda tup: tup[0])
        words_and_frequencies.sort(key=lambda tup: tup[1], reverse=True)
        for word, freq in words_and_frequencies:
            if len(self.itos) == self.size:
                break
            self.itos.append(word)
        if self.size is None:
            self.size = len(self.itos)
        self.stoi = self._make_reverse_mapping(self.itos)

    def save(self, fname: str) -> None:
        """Saves a vocabulary in json format

        Parameter
        ---------
        fname: str
            Filename to save vocabulary
        """
        vocab = {'stoi': self.stoi, 'itos': self.itos, 'size': self.size}
        with open(fname, 'w') as f:
            json.dump(vocab, f)

    @classmethod
    def load(cls, fname: str) -> 'GroverAtomVocabularyBuilder':
        """Loads vocabulary from the specified json file

        Parameters
        ----------
        fname: str
            JSON file containing vocabulary

        Returns
        -------
        vocab: GroverAtomVocabularyBuilder
            A grover atom vocabulary builder which can be used for encoding
        """
        with open(fname, 'r') as f:
            data = json.load(f)
        vocab = cls()
        vocab.stoi, vocab.itos, vocab.size = data['stoi'], data['itos'], data[
            'size']
        return vocab

    @staticmethod
    def atom_to_vocab(mol: RDKitMol, atom: RDKitAtom) -> str:
        """Convert atom to vocabulary.

        Parameters
        ----------
        mol: RDKitMol
            an molecule object
        atom: RDKitAtom
            the target atom.

        Returns
        -------
        vocab: str
            The generated atom vocabulary with its contexts.

        Example
        -------
        >>> from rdkit import Chem
        >>> mol = Chem.MolFromSmiles('[C@@H](C)C(=O)O')
        >>> GroverAtomVocabularyBuilder.atom_to_vocab(mol, mol.GetAtomWithIdx(0))
        'C_C-SINGLE2'
        >>> GroverAtomVocabularyBuilder.atom_to_vocab(mol, mol.GetAtomWithIdx(3))
        'O_C-DOUBLE1'
        """
        atom_neighbors: Dict[str, int] = Counter()
        for a in atom.GetNeighbors():
            bond = mol.GetBondBetweenAtoms(atom.GetIdx(), a.GetIdx())
            atom_neighbors[str(a.GetSymbol()) + "-" +
                           str(bond.GetBondType())] += 1
        keys = list(atom_neighbors.keys())
        # sorting the atoms neighbors
        keys.sort()
        output = atom.GetSymbol()
        # concatenating the sorted neighbors
        for key in keys:
            output = "%s_%s%d" % (output, key, atom_neighbors[key])
        return output

    def _make_reverse_mapping(self, itos):
        return {tok: i for i, tok in enumerate(itos)}

    def encode(self, mol: RDKitMol, atom: RDKitAtom) -> str:
        """Encodes an atom in a molecule

        Parameter
        ---------
        mol: RDKitMol
            An RDKitMol object
        atom: RDKitAtom
            An atom in the molecule

        Returns
        -------
        vocab: str
            The vocabulary of the atom in the molecule.
        """
        return self.stoi.get(self.atom_to_vocab(mol, atom))


class GroverBondVocabularyBuilder(VocabularyBuilder):
    """Bond Vocabulary Builder for Grover

    This module can be used to generate bond vocabulary from SMILES strings
    for the GROVER pretraining task.

    For assigning the vocabulary of a bond, we consider the features of the bond
    and the context of the bond. The context of bond is the feature of the bond under
    consideration and the feature of the bonds of atom in which the bond begins and ends.
    It is formed by the concatenation of atomSymbol-bondFeature-Count where atomSymbol
    is the symbol of neighboring atom, bondFeature is the type of bond and count is the
    number of such atomSymbol-bondFeature pairs in the surrounding context.

    The feature of a bond is determined by three sub-features: the type of bond (single or double bond),
    the RDKit StereoConfiguration of the bond and RDKit BondDir. For the C-C bond
    in CCC, the type of bond is SINGLE, its stereo is NONE and the bond does not have
    direction. Hence, the feature of the bond is SINGLE-STEREONONE-NONE.

    For assigning the vocabulary, we should also have to look at the neighboring bonds.
    Consider the molecule 'CC(=O)C'. It has three bonds. The C-C bond has two neighbors.
    The first C atom has no other bonds, so it contributes no context. The second C atom
    has one bond with an O atom and one bond with a C atom. Consider the C=O double bond.
    The bond feature is DOUBLE-STEREONONE-NONE. The corresponding context is
    atomSymbol-bondFeature-Count. This gives us C-(DOUBLE-STEREONONE-NONE)1.
    Similary, it also has another bond with a C atom which gives the
    context C-(SINGLE-STEREONONE-NONE)1. Hence, the vocabulary of
    the bond is '(SINGLE-STEREONONE-NONE)_C-(DOUBLE-STEREONONE-NONE)1_C-(SINGLE-STEREONONE-NONE)1'

    The algorithm enumerates vocabulary of all bonds in the dataset and makes a vocabulary to
    index mapping by sorting the vocabulary by frequency and then alphabetically. The `max_size`
    parameter can be used for setting the size of the vocabulary. When this parameter is set,
    the algorithm stops adding new words to the index when the vocabulary size reaches `max_size`.

    Parameters
    ----------
    max_size: int (optional)
        Maximum size of vocabulary

    Example
    -------
    >>> import tempfile
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> file = tempfile.NamedTemporaryFile()
    >>> dataset = dc.data.NumpyDataset(X=[['CCC']])
    >>> vocab = GroverBondVocabularyBuilder()
    >>> vocab.build(dataset)
    >>> vocab.stoi
    {'<pad>': 0, '<other>': 1, '(SINGLE-STEREONONE-NONE)_C-(SINGLE-STEREONONE-NONE)1': 2}
    >>> vocab.save(file.name)
    >>> loaded_vocab = GroverBondVocabularyBuilder.load(file.name)
    >>> mol = Chem.MolFromSmiles('CCC')
    >>> loaded_vocab.encode(mol, mol.GetBondWithIdx(0))
    2

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """
    BOND_FEATURES = ['BondType', 'Stereo', 'BondDir']

    def __init__(self, max_size: Optional[int] = None):
        self.specials = ('<pad>', '<other>')
        self.min_freq = 1
        self.size = max_size
        self.itos = list(self.specials)
        self.stoi = self._make_reverse_mapping(self.itos)
        self.pad_index = 0
        self.other_index = 1

    def build(self, dataset: Dataset) -> None:
        """Builds vocabulary

        Parameters
        ----------
        dataset: dc.data.Dataset
            A dataset object with SMILEs strings in X attribute.
        """
        counter: Dict[str, int] = Counter()
        for x, _, _, _ in dataset.itersamples():
            smiles = x[0]
            mol = Chem.MolFromSmiles(smiles)
            for bond in mol.GetBonds():
                v = self.bond_to_vocab(mol, bond)
                counter[v] += 1

        # sort first by frequency, then alphabetically
        words_and_frequencies = sorted(counter.items(), key=lambda tup: tup[0])
        words_and_frequencies.sort(key=lambda tup: tup[1], reverse=True)
        for word, freq in words_and_frequencies:
            if len(self.itos) == self.size:
                break
            self.itos.append(word)
        if self.size is None:
            self.size = len(self.itos)
        self.stoi = self._make_reverse_mapping(self.itos)

    def save(self, fname: str) -> None:
        """Saves a vocabulary in json format

        Parameter
        ---------
        fname: str
            Filename to save vocabulary
        """
        vocab = {'stoi': self.stoi, 'itos': self.itos, 'size': self.size}
        with open(fname, 'w') as f:
            json.dump(vocab, f)

    @classmethod
    def load(cls, fname: str) -> 'GroverBondVocabularyBuilder':
        """Loads vocabulary from the specified json file

        Parameters
        ----------
        fname: str
            JSON file containing vocabulary

        Returns
        -------
        vocab: GroverAtomVocabularyBuilder
            A grover atom vocabulary builder which can be used for encoding
        """
        with open(fname, 'r') as f:
            data = json.load(f)
        vocab = cls()
        vocab.stoi, vocab.itos, vocab.size = data['stoi'], data['itos'], data[
            'size']
        return vocab

    @staticmethod
    def bond_to_vocab(mol: RDKitMol, bond: RDKitBond):
        """Convert bond to vocabulary.

        The algorithm considers only one-hop neighbor atoms.

        Parameters
        ----------
        mol: RDKitMole
            the molecule object
        bond: RDKitBond
            the target bond

        Returns
        -------
        vocab: str
            the generated bond vocabulary with its contexts.

        Example
        -------
        >>> from rdkit import Chem
        >>> mol = Chem.MolFromSmiles('[C@@H](C)C(=O)O')
        >>> GroverBondVocabularyBuilder.bond_to_vocab(mol, mol.GetBondWithIdx(0))
        '(SINGLE-STEREONONE-NONE)_C-(SINGLE-STEREONONE-NONE)1'
        >>> GroverBondVocabularyBuilder.bond_to_vocab(mol, mol.GetBondWithIdx(2))
        '(DOUBLE-STEREONONE-NONE)_C-(SINGLE-STEREONONE-NONE)2'
        """
        bond_neighbors: Dict[str, int] = Counter()
        two_neighbors = (bond.GetBeginAtom(), bond.GetEndAtom())
        two_indices = [a.GetIdx() for a in two_neighbors]
        for nei_atom in two_neighbors:
            for a in nei_atom.GetNeighbors():
                a_idx = a.GetIdx()
                if a_idx in two_indices:
                    continue
                tmp_bond = mol.GetBondBetweenAtoms(nei_atom.GetIdx(), a_idx)
                bond_neighbors[str(nei_atom.GetSymbol()) + '-' +
                               GroverBondVocabularyBuilder.
                               _get_bond_feature_name(tmp_bond)] += 1
        keys = list(bond_neighbors.keys())
        keys.sort()
        output = GroverBondVocabularyBuilder._get_bond_feature_name(bond)
        for k in keys:
            output = "%s_%s%d" % (output, k, bond_neighbors[k])
        return output

    @staticmethod
    def _get_bond_feature_name(bond: RDKitBond):
        """Return the string format of bond features."""
        ret = []
        for bond_feature in GroverBondVocabularyBuilder.BOND_FEATURES:
            fea = eval(f"bond.Get{bond_feature}")()
            ret.append(str(fea))
        return '(' + '-'.join(ret) + ')'

    def _make_reverse_mapping(self, itos):
        return {tok: i for i, tok in enumerate(itos)}

    def encode(self, mol: RDKitMol, bond: RDKitBond) -> str:
        """Encodes a bond in a molecule

        Parameter
        ---------
        mol: RDKitMol
            An RDKitMol object
        bond: RDKitBond
            A bond in the molecule

        Returns
        -------
        vocab: str
            The vocabulary of the bond in the molecule.
        """
        return self.stoi.get(self.bond_to_vocab(mol, bond))


class GroverAtomVocabTokenizer(Featurizer):
    """Grover Atom Vocabulary Tokenizer

    The Grover Atom vocab tokenizer is used for tokenizing an atom using a
    vocabulary generated by GroverAtomVocabularyBuilder.

    Example
    -------
    >>> import tempfile
    >>> import deepchem as dc
    >>> from deepchem.feat.vocabulary_builders.grover_vocab import GroverAtomVocabularyBuilder
    >>> file = tempfile.NamedTemporaryFile()
    >>> dataset = dc.data.NumpyDataset(X=[['CC(=O)C', 'CCC']])
    >>> vocab = GroverAtomVocabularyBuilder()
    >>> vocab.build(dataset)
    >>> vocab.save(file.name)  # build and save the vocabulary
    >>> atom_tokenizer = GroverAtomVocabTokenizer(file.name)
    >>> mol = Chem.MolFromSmiles('CC(=O)C')
    >>> atom_tokenizer.featurize([(mol, mol.GetAtomWithIdx(0))])[0]
    2

    Parameters
    ----------
    fname: str
        Filename of vocabulary generated by GroverAtomVocabularyBuilder
    """

    def __init__(self, fname: str):
        self.vocabulary = GroverAtomVocabularyBuilder.load(fname)

    def _featurize(self, datapoint):
        return self.vocabulary.encode(datapoint[0], datapoint[1])


class GroverBondVocabTokenizer(Featurizer):
    """Grover Bond Vocabulary Tokenizer

    The Grover Bond vocab tokenizer is used for tokenizing a bond using a
    vocabulary generated by GroverBondVocabularyBuilder.

    Example
    -------
    >>> import tempfile
    >>> import deepchem as dc
    >>> from deepchem.feat.vocabulary_builders.grover_vocab import GroverBondVocabularyBuilder
    >>> file = tempfile.NamedTemporaryFile()
    >>> dataset = dc.data.NumpyDataset(X=[['CC(=O)C', 'CCC']])
    >>> vocab = GroverBondVocabularyBuilder()
    >>> vocab.build(dataset)
    >>> vocab.save(file.name)  # build and save the vocabulary
    >>> bond_tokenizer = GroverBondVocabTokenizer(file.name)
    >>> mol = Chem.MolFromSmiles('CC(=O)C')
    >>> bond_tokenizer.featurize([(mol, mol.GetBondWithIdx(0))])[0]
    2

    Parameters
    ----------
    fname: str
        Filename of vocabulary generated by GroverAtomVocabularyBuilder
    """

    def __init__(self, fname: str):
        self.vocabulary = GroverBondVocabularyBuilder.load(fname)

    def _featurize(self, datapoint):
        return self.vocabulary.encode(datapoint[0], datapoint[1])



"""
Tests that sequence handling utilities work.
"""
__author__ = "Bharath Ramsundar"
__license__ = "MIT"

import unittest
import os

import numpy as np

import deepchem as dc

LETTERS = "XYZ"


class TestSeq(unittest.TestCase):
    """
    Tests sequence handling utilities.
    """

    def setUp(self):
        super(TestSeq, self).setUp()
        self.current_dir = os.path.dirname(os.path.abspath(__file__))

    def test_one_hot_simple(self):
        sequences = np.array(["ACGT", "GATA", "CGCG"])
        sequences = dc.utils.genomics_utils.seq_one_hot_encode(sequences)
        self.assertEqual(sequences.shape, (3, 5, 4, 1))

    def test_one_hot_mismatch(self):
        # One sequence has length longer than others. This should throw a
        # ValueError.

        with self.assertRaises(ValueError):
            sequences = np.array(["ACGTA", "GATA", "CGCG"])
            sequences = dc.utils.genomics_utils.seq_one_hot_encode(sequences)

    def test_encode_fasta_sequence(self):
        # Test it's possible to load a sequence with an aribrary alphabet from a fasta file.
        fname = os.path.join(self.current_dir, "./assets/example.fasta")

        encoded_seqs = dc.utils.genomics_utils.encode_bio_sequence(
            fname, letters=LETTERS)
        expected = np.expand_dims(
            np.array([
                [[1, 0], [0, 1], [0, 0]],
                [[0, 1], [0, 0], [1, 0]],
            ]), -1)

        np.testing.assert_array_equal(expected, encoded_seqs)

    def test_encode_fastq_sequence(self):
        fname = os.path.join(self.current_dir, "./assets/example.fastq")

        encoded_seqs = dc.utils.genomics_utils.encode_bio_sequence(
            fname, file_type="fastq", letters=LETTERS)

        expected = np.expand_dims(
            np.array([
                [[1, 0], [0, 1], [0, 0]],
                [[0, 1], [0, 0], [1, 0]],
            ]), -1)

        np.testing.assert_array_equal(expected, encoded_seqs)



"""
Genomic data handling Iterable.
"""
from typing import Dict, Iterable, Union
import numpy as np


def seq_one_hot_encode(sequences, letters: str = 'ATCGN') -> np.ndarray:
    """One hot encodes list of genomic sequences.

    Sequences encoded have shape (N_sequences, N_letters, sequence_length, 1).
    These sequences will be processed as images with one color channel.

    Parameters
    ----------
    sequences: np.ndarray or Iterator[Bio.SeqRecord]
        Iterable object of genetic sequences
    letters: str, optional (default "ATCGN")
        String with the set of possible letters in the sequences.

    Raises
    ------
    ValueError:
        If sequences are of different lengths.

    Returns
    -------
    np.ndarray
        A numpy array of shape `(N_sequences, N_letters, sequence_length, 1)`.
    """

    # The label encoder is given characters for ACGTN
    letter_encoder = {l: i for i, l in enumerate(letters)}
    alphabet_length = len(letter_encoder)

    # Peak at the first sequence to get the length of the sequence.
    if isinstance(sequences, np.ndarray):
        first_seq = sequences[0]
        tail_seq = sequences[1:]
    else:
        first_seq = next(sequences)
        tail_seq = sequences

    sequence_length = len(first_seq)
    seqs = []
    seqs.append(
        _seq_to_encoded(first_seq, letter_encoder, alphabet_length,
                        sequence_length))

    for other_seq in tail_seq:
        if len(other_seq) != sequence_length:
            raise ValueError("The genetic sequences must have a same length")
        seqs.append(
            _seq_to_encoded(other_seq, letter_encoder, alphabet_length,
                            sequence_length))

    return np.expand_dims(np.array(seqs), -1)


def _seq_to_encoded(seq: Union[str, Iterable[str]], letter_encoder: Dict[str,
                                                                         int],
                    alphabet_length: int, sequence_length: int) -> np.ndarray:
    """One hot encodes a genomic sequence.

    Sequences encoded have shape (N_sequences, N_letters, sequence_length, 1).
    These sequences will be processed as images with one color channel.

    Parameters
    ----------
    seq: str or Bio.SeqRecord
        a genetic sequence
    letter_encoder: Dict[str, int]
        The keys are letters and the values are unique int values (like 0, 1, 2...).
    alphabet_length: int
        Length with the set of possible letters in the sequences.
    sequence_length: int
        Length with a genetic sequence

    Returns
    -------
    encoded_seq: np.ndarray
        A numpy array of shape `(N_letters, sequence_length)`.
    """
    encoded_seq = np.zeros((alphabet_length, sequence_length))
    seq_ints = [letter_encoder[s] for s in seq]
    encoded_seq[seq_ints, np.arange(sequence_length)] = 1
    return encoded_seq


def encode_bio_sequence(fname: str,
                        file_type: str = "fasta",
                        letters: str = "ATCGN") -> np.ndarray:
    """
    Loads a sequence file and returns an array of one-hot sequences.

    Parameters
    ----------
    fname: str
        Filename of fasta file.
    file_type: str, optional (default "fasta")
        The type of file encoding to process, e.g. fasta or fastq, this
        is passed to Biopython.SeqIO.parse.
    letters: str, optional (default "ATCGN")
        The set of letters that the sequences consist of, e.g. ATCG.

    Returns
    -------
    np.ndarray
        A numpy array of shape `(N_sequences, N_letters, sequence_length, 1)`.

    Notes
    -----
    This function requires BioPython to be installed.
    """
    try:
        from Bio import SeqIO
    except ModuleNotFoundError:
        raise ImportError("This function requires BioPython to be installed.")

    sequences = SeqIO.parse(fname, file_type)
    return seq_one_hot_encode(sequences, letters)



import time
import logging
import os
from collections.abc import Sequence as SequenceCollection
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union, Sequence
import torch
import torch.nn as nn
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.optimizers import LearningRateSchedule
from deepchem.utils.typing import LossFn, OneOrMany

logger = logging.getLogger(__name__)


class ModularTorchModel(TorchModel):
    """ModularTorchModel is a subclass of TorchModel that allows for components to be
    pretrained and then combined into a final model. It is designed to be subclassed
    for specific models and is not intended to be used directly. There are 3 main differences
    between ModularTorchModel and TorchModel:

    - The build_components() method is used to define the components of the model.
    - The components are combined into a final model with the build_model() method.
    - The loss function is defined with the loss_func method. This may access the
      components to compute the loss using intermediate values from the network, rather
      than just the full forward pass output.

    Here is an example of how to use ModularTorchModel to pretrain a linear layer, load
    it into another network and then finetune that network:

    >>> import numpy as np
    >>> import deepchem as dc
    >>> import torch
    >>> n_samples = 6
    >>> n_feat = 3
    >>> n_hidden = 2
    >>> n_tasks = 6
    >>> pt_tasks = 3
    >>> X = np.random.rand(n_samples, n_feat)
    >>> y_pretrain = np.zeros((n_samples, pt_tasks)).astype(np.float32)
    >>> dataset_pt = dc.data.NumpyDataset(X, y_pretrain)
    >>> y_finetune = np.zeros((n_samples, n_tasks)).astype(np.float32)
    >>> dataset_ft = dc.data.NumpyDataset(X, y_finetune)
    >>> components = {'linear': torch.nn.Linear(n_feat, n_hidden),
    ... 'activation': torch.nn.ReLU(), 'head': torch.nn.Linear(n_hidden, n_tasks)}
    >>> model = torch.nn.Sequential(components['linear'], components['activation'],
    ... components['head'])
    >>> modular_model = dc.models.torch_models.modular.ModularTorchModel(model, components)
    >>> def example_loss_func(inputs, labels, weights):
    ...    return (torch.nn.functional.mse_loss(model(inputs), labels[0]) * weights[0]).mean()
    >>> modular_model.loss_func = example_loss_func
    >>> def example_model_build():
    ...     return torch.nn.Sequential(components['linear'], components['activation'],
    ... components['head'])
    >>> modular_model.build_model = example_model_build
    >>> pretrain_components = {'linear': torch.nn.Linear(n_feat, n_hidden),
    ... 'activation': torch.nn.ReLU(), 'head': torch.nn.Linear(n_hidden, pt_tasks)}
    >>> pretrain_model = torch.nn.Sequential(pretrain_components['linear'],
    ... pretrain_components['activation'], pretrain_components['head'])
    >>> pretrain_modular_model = dc.models.torch_models.modular.ModularTorchModel(pretrain_model,
    ... pretrain_components)
    >>> def example_pt_loss_func(inputs, labels, weights):
    ...     return (torch.nn.functional.mse_loss(pretrain_model(inputs), labels[0]) * weights[0]).mean()
    >>> pretrain_modular_model.loss_func = example_pt_loss_func
    >>> pt_loss = pretrain_modular_model.fit(dataset_pt, nb_epoch=1)
    >>> modular_model.load_from_pretrained(pretrain_modular_model, components=['linear'])
    >>> ft_loss = modular_model.fit(dataset_ft, nb_epoch=1)

    """

    def __init__(self, model: nn.Module, components: dict, **kwargs):
        """Create a ModularTorchModel.

        Parameters
        ----------
        model: nn.Module
            The model to be trained.
        components: dict
            A dictionary of the components of the model. The keys are the names of the
            components and the values are the components themselves.
        """

        self.model = model
        self.components = components
        # FIXME self.loss_func is an incorrect argument for TorchModel.loss because
        # it performs more than computing loss
        super().__init__(self.model, self.loss_func, **kwargs)
        self.model.to(self.device)
        self.components = {
            k: v.to(self.device)
            for k, v in self.components.items()
            if isinstance(v, nn.Module)
        }

    def build_model(self) -> nn.Module:
        """Builds the final model from the components."""
        raise NotImplementedError("Subclass must define the components")

    def build_components(self) -> dict:
        """Creates the components dictionary, with the keys being the names of the
        components and the values being torch.nn.module objects."""
        raise NotImplementedError("Subclass must define the components")

    def loss_func(self, inputs: OneOrMany[torch.Tensor], labels: Sequence,
                  weights: Sequence) -> torch.Tensor:
        """Defines the loss function for the model which can access the components
        using self.components. The loss function should take the inputs, labels, and
        weights as arguments and return the loss."""
        raise NotImplementedError("Subclass must define the loss function")

    def freeze_components(self, components: List[str]):
        """Freezes or unfreezes the parameters of the specified components.

        Components string refers to keys in self.components.

        Parameters
        ----------
        components: List[str]
            The components to freeze.
        """
        for component in components:
            for param in self.components[component].parameters():
                param.requires_grad = False

    def unfreeze_components(self, components: List[str]):
        """Unfreezes the parameters of the specified components.

        Components string refers to keys in self.components.

        Parameters
        ----------
        components: List[str]
            The components to unfreeze.
        """
        for component in components:
            for param in self.components[component].parameters():
                param.requires_grad = True

    def fit_generator(self,
                      generator: Iterable[Tuple[Any, Any, Any]],
                      max_checkpoints_to_keep: int = 5,
                      checkpoint_interval: int = 1000,
                      restore: bool = False,
                      variables: Optional[List[torch.nn.Parameter]] = None,
                      loss: Optional[LossFn] = None,
                      callbacks: Union[Callable, List[Callable]] = [],
                      all_losses: Optional[List[float]] = None) -> float:
        """Train this model on data from a generator. This method is similar to
        the TorchModel implementation, but it passes the inputs directly to the
        loss function, rather than passing them through the model first.  This
        enables the loss to be calculated from intermediate steps of the model
        and not just the final output.

        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in training steps.
            Set this to 0 to disable automatic checkpointing.
        restore: bool
            if True, restore the model from the most recent checkpoint and continue training
            from there.  If False, retrain the model from scratch.
        variables: list of torch.nn.Parameter
            the variables to train.  If None (the default), all trainable variables in
            the model are used.
        loss: function
            a function of the form f(outputs, labels, weights) that computes the loss
            for each batch.  If None (the default), the model's standard loss function
            is used.
        callbacks: function or list of functions
            one or more functions of the form f(model, step) that will be invoked after
            every step.  This can be used to perform validation, logging, etc.
        all_losses: Optional[List[float]], optional (default None)
            If specified, all logged losses are appended into this list. Note that
            you can call `fit()` repeatedly with the same list and losses will
            continue to be appended.

        Returns
        -------
        The average loss over the most recent checkpoint interval
        """

        if not isinstance(callbacks, SequenceCollection):
            callbacks = [callbacks]
        self._ensure_built()
        self.model.train()
        avg_loss = 0.0
        last_avg_loss = 0.0
        averaged_batches = 0
        # FIXME This line is not needed as loss is computed inside the call to loss_func
        if loss is None:
            loss = self._loss_fn
        if variables is None:
            optimizer = self._pytorch_optimizer
            lr_schedule = self._lr_schedule
        else:
            var_key = tuple(variables)
            if var_key in self._optimizer_for_vars:
                optimizer, lr_schedule = self._optimizer_for_vars[var_key]
            else:
                optimizer = self.optimizer._create_pytorch_optimizer(variables)
                if isinstance(self.optimizer.learning_rate,
                              LearningRateSchedule):
                    lr_schedule = self.optimizer.learning_rate._create_pytorch_schedule(
                        optimizer)
                else:
                    lr_schedule = None
                self._optimizer_for_vars[var_key] = (optimizer, lr_schedule)
        time1 = time.time()

        # Main training loop.

        for batch in generator:
            if restore:
                self.restore()
                restore = False
            inputs: OneOrMany[torch.Tensor]
            inputs, labels, weights = self._prepare_batch(batch)

            # Execute the loss function, accumulating the gradients.

            if isinstance(inputs, list) and len(inputs) == 1:
                inputs = inputs[0]

            optimizer.zero_grad()
            batch_loss = self.loss_func(inputs, labels, weights)
            batch_loss.backward()
            optimizer.step()
            if lr_schedule is not None:
                lr_schedule.step()
            self._global_step += 1
            current_step = self._global_step

            avg_loss += float(batch_loss)

            # Report progress and write checkpoints.
            averaged_batches += 1
            should_log = (current_step % self.log_frequency == 0)
            if should_log:
                avg_loss = float(avg_loss) / averaged_batches
                logger.info('Ending global_step %d: Average loss %g' %
                            (current_step, avg_loss))
                if all_losses is not None:
                    all_losses.append(avg_loss)
                # Capture the last avg_loss in case of return since we're resetting to 0 now
                last_avg_loss = avg_loss
                avg_loss = 0.0
                averaged_batches = 0

            if checkpoint_interval > 0 and current_step % checkpoint_interval == checkpoint_interval - 1:
                self.save_checkpoint(max_checkpoints_to_keep)
            for c in callbacks:
                c(self, current_step)
            if self.tensorboard and should_log:
                self._log_scalar_to_tensorboard('loss', batch_loss,
                                                current_step)
            if (self.wandb_logger is not None) and should_log:
                all_data = dict({'train/loss': batch_loss})
                self.wandb_logger.log_data(all_data, step=current_step)

        # Report final results.
        if averaged_batches > 0:
            avg_loss = float(avg_loss) / averaged_batches
            logger.info('Ending global_step %d: Average loss %g' %
                        (current_step, avg_loss))
            if all_losses is not None:
                all_losses.append(avg_loss)
            last_avg_loss = avg_loss

        if checkpoint_interval > 0:
            self.save_checkpoint(max_checkpoints_to_keep)

        time2 = time.time()
        logger.info("TIMING: model fitting took %0.3f s" % (time2 - time1))
        return last_avg_loss

    def load_from_pretrained(  # type: ignore
            self,
            source_model: Optional["ModularTorchModel"] = None,
            components: Optional[List[str]] = None,
            checkpoint: Optional[str] = None,
            model_dir: Optional[str] = None,
            inputs: Optional[Sequence[Any]] = None,
            **kwargs) -> None:
        """Copies parameter values from a pretrained model. The pretrained model can be loaded as a source_model (ModularTorchModel object), checkpoint (pytorch .ckpt file) or a model_dir (directory with .ckpt files).
        Specific components can be chosen by passing a list of strings with the desired component names. If both a source_model and a checkpoint/model_dir are loaded, the source_model weights will be loaded.

        Parameters
        ----------
        source_model: dc.ModularTorchModel, required
            source_model can either be the pretrained model or a dc.TorchModel with
            the same architecture as the pretrained model. It is used to restore from
            a checkpoint, if value_map is None and to create a default assignment map
            if assignment_map is None
        checkpoint: str, default None
            the path to the checkpoint file to load.  If this is None, the most recent
            checkpoint will be chosen automatically.  Call get_checkpoints() to get a
            list of all available checkpoints
        model_dir: str, default None
            Restore source model from custom model directory if needed
        inputs: List, input tensors for model
            if not None, then the weights are built for both the source and self.
        """
        if inputs is not None:
            # Ensure weights for both models are built.
            if source_model:
                source_model.model(inputs)
            self.model(inputs)

        self._ensure_built()

        if source_model is not None:
            for name, module in source_model.components.items():
                if components is None or name in components:
                    self.components[name].load_state_dict(module.state_dict())
            self.build_model()

        elif source_model is None:
            self.restore(components=components,
                         checkpoint=checkpoint,
                         model_dir=model_dir)

    def save_checkpoint(self, max_checkpoints_to_keep=5, model_dir=None):
        """
        Saves the current state of the model and its components as a checkpoint file in the specified model directory.
        It maintains a maximum number of checkpoint files, deleting the oldest one when the limit is reached.

        Parameters
        ----------
        max_checkpoints_to_keep: int, default 5
            Maximum number of checkpoint files to keep.
        model_dir: str, default None
            The directory to save the checkpoint file in. If None, the model_dir specified in the constructor is used.
        """

        if model_dir is None:
            model_dir = self.model_dir
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        data = {
            'model': self.model.state_dict(),
            'optimizer_state_dict': self._pytorch_optimizer.state_dict(),
            'global_step': self._global_step
        }

        for name, component in self.components.items():
            data[name] = component.state_dict()

        temp_file = os.path.join(model_dir, 'temp_checkpoint.pt')
        torch.save(data, temp_file)

        # Rename and delete older files.

        paths = [
            os.path.join(model_dir, 'checkpoint%d.pt' % (i + 1))
            for i in range(max_checkpoints_to_keep)
        ]
        if os.path.exists(paths[-1]):
            os.remove(paths[-1])
        for i in reversed(range(max_checkpoints_to_keep - 1)):
            if os.path.exists(paths[i]):
                os.rename(paths[i], paths[i + 1])
        os.rename(temp_file, paths[0])

    def restore(  # type: ignore
            self,
            components: Optional[List[str]] = None,
            checkpoint: Optional[str] = None,
            model_dir: Optional[str] = None) -> None:
        """
        Restores the state of a ModularTorchModel from a checkpoint file.

        If no checkpoint file is provided, it will use the latest checkpoint found in the model directory. If a list of component names is provided, only the state of those components will be restored.

        Parameters
        ----------
        components: Optional[List[str]]
            A list of component names to restore. If None, all components will be restored.
        checkpoint: Optional[str]
            The path to the checkpoint file. If None, the latest checkpoint in the model directory will
            be used.
        model_dir: Optional[str]
            The path to the model directory. If None, the model directory used to initialize the model will be used.
        """
        if checkpoint is None:
            checkpoints = sorted(self.get_checkpoints(model_dir))
            if len(checkpoints) == 0:
                raise ValueError('No checkpoint found')
            checkpoint = checkpoints[0]
        data = torch.load(checkpoint)
        for name, state_dict in data.items():
            if name != 'model' and name in self.components.keys():
                if components is None or name in components:
                    self.components[name].load_state_dict(state_dict)

        self.build_model()
        self._pytorch_optimizer.load_state_dict(data['optimizer_state_dict'])
        self._global_step = data['global_step']
